#################
## LCA Machine ##
#################

norm.density <- function(Y, mean, sigma){
  if (is.null(dim(Y)) == TRUE){
    N <- 1
    P <- length(Y)
    Y <- matrix(Y, N, P)
  } else {
    N <- dim(Y)[1]
    P <- dim(Y)[2]
  }
  mu <- matrix(mean, N, P, byrow = TRUE)
  sigma <- sigma
  tmp <- exp( -0.5*(Y - mu) %*% solve(sigma) %*% t(Y - mu))
  tmp <- (det(sigma)^(-0.5)) * diag(tmp)* (1/sqrt(2*pi))^P
  return(tmp)
}

t.density <- function(n, mean, sigma, df = 4){ 
  ## Requires R-package "mvtnorm"
  n <- n 
  P <- length(mean)
  df <- df 
  dat <- t(t(rmvnorm(n, mean = rep(0, P), sigma) * sqrt(df / rchisq(n, df))) + mean)
  return(dat)
}

unif.density <- function(n, mean, sigma, df = 4){ 
  ## Requires R-package "mvtnorm"
  n <- n 
  P <- length(mean)
  df <- df 
  dat <- t(t(rmvnorm(n, mean = rep(0, P), sigma) * sqrt(df / rchisq(n, df))) + mean)
  return(dat)
}

boot.density <- function(test.obs, dat, param, nboot = 100, alpha = 0.05){
  n.tr <- dim(dat)[1]
  P <- dim(dat)[2]
  train.center <- param$mu
  train.cov <- param$cov
  ncls <- dim(param$mu)[2]
  nboot <- nboot
  alpha <- alpha
  test.result <- rep(0,ncls)
  for (cls in 1:ncls){ 
    tmp.tr <- norm.density(dat, mean = train.center[,cls], sigma = train.cov[,,cls])
    tmp.new <- norm.density(test.obs, mean = train.center[,cls], sigma = train.cov[,,cls])
    tstat.obs <- mean(tmp.tr) - tmp.new
    tmp.boot <- sample(tmp.tr, n.tr*nboot, replace = TRUE)
    boot.dist <- matrix(tmp.boot, nboot, n.tr,  byrow = TRUE)
    boot.dist <- apply(boot.dist, 2, mean)
    tstat.boot <- boot.dist - tmp.new 
    test.result[cls] <- sum(abs(tstat.boot) >= abs(tstat.obs)) / n.tr
  }
  out <- prod(test.result <= alpha)  
  return(out)
}  

LCA.machine <- function(Y.train, Y.test, X.train, X.test, mod=NULL, param, model, constraint, alpha = 0.05){
  Y.train <- Y.train
  Y.test <- Y.test
  if (is.null(dim(Y.test)) == TRUE){
    test <- matrix(Y.test, 1, model$nci)
    n <- 1
  } else {
    n <- dim(Y.test)[1] 
  }
  npat <- prod(model$nrc)
  
  # Train the model #
  if (is.null(mod) == TRUE){
    M0 <- LCA.EM.iteration(Y.train, X=X.train, param, model, constraint)
  } else {
    M0 <- mod
  }
  
  # Find the null distribution #
  Y.pat <- matrix(0,npat,model$nci)
  for (p in 1:model$nci){
    Y.pat[,p] <- rep(1:model$nrc[p], 2^(p-1), each = 2^(model$nci-p))  
  }
  null.dist <- apply(Y.pat, MARGIN = 1, function(Y.pat){LCA.loglikf(Y.pat, R=NULL, X=X.train, param = M0$param, model = model)}) 
  null.prob <- exp(null.dist) / sum(exp(null.dist)) # Distribution of measure under M0
  
  # Calculate test statistic for testing data #
  new.dist <- apply(Y.test, MARGIN = 1, function(Y.test){LCA.loglikf(Y.test, R=NULL, X=NULL, param = M0$param, model = model)})
  stat <- rep(NA, n) 
  for (i in 1:n){
    stat[i] <- sum(null.prob[null.dist <= new.dist[i]]) 
  }
  
  # Calculate Rejection Region #
  nlist <- which(cumsum(null.prob[order(null.prob)]) > alpha)[1] - 1
  if (nlist > 0){
    RR <- Y.pat[rank(null.prob) <= nlist,]
  } else {
    RR <- NULL
  }
  output <- list()
  output$stat <- stat # test statistic for each testing data
  output$RR <- RR # Rejection Region
  output$M0 <- M0 # training model
  return(output)
}

convert.pat <- function(vector,K){
  n <- length(vector)
  A <- 0
  if (is.null(K) == TRUE){
    K <- max(vector) + 1
  }
  for (i in 1:n){
    A <- A + vector[i]*K^i  
  }
  return(A)
} 

pred.comparison <- function(Y.train, Y.test, X.train, X.test, param, model){
  # Input: Training, Testing data #
  Y.train <- Y.train
  Y.test <- Y.test
  X.train <- X.train
  X.test <- X.test
  param <- param 
  model <- model
  n.tr <- dim(Y.train)[1]
  n.tst <- dim(Y.test)[1]
  constraint <- list()
  constraint$phi <- array(NA, c(model$nci, max(model$nrc), model$nc))
  constraint$gamma <- rep(NA, model$nc)
  
  # Train the model #
  M0 <- LCA.EM.iteration(Y.train, X.train, param, model, constraint)
  
  # Case I : make prediction based on conventional method #
  tmp <- LCA.Post(Y.test, R=NULL, X.test, M0$param, model, constraint)
  pred1 <- apply(tmp$post.D, 1, which.max)
  
  # Case II : make prediction based on proposed method #
  output <- LCA.machine(Y.train, Y.test, X.train, X.test, mod=M0, param, model, constraint, alpha = 0.05)
  test.id <- which(output$stat <= 0.05)
  tmp <- LCA.Post(Y.test[-test.id,], R=NULL, X.test[-test.id,], M0$param, model, constraint)
  pred2 <- numeric(n.tst)
  pred2[-test.id] <- apply(tmp$post.D, 1, which.max)
  pred2[test.id] <- 99
  
  output <- list()
  output$pred1 <- pred1
  output$pred2 <- pred2
  return(output)
}


Rand.index <- function (group1, group2) {
  x <- abs(sapply(group1, function(x) x - group1))
  x[x > 1] <- 1
  y <- abs(sapply(group2, function(x) x - group2))
  y[y > 1] <- 1
  sg <- sum(abs(x - y))/2
  bc <- choose(dim(x)[1], 2)
  ri <- 1 - sg/bc
  return(ri)
}

miscls.rate <- function(obs, true){
  obs <- obs
  true <- true
  N <- length(obs)
  if (length(true) != N){
    cat('Error : the length of two objects are different','\n')
  } else {
    tmp <- sum(true != obs) / N 
    return(tmp)
  }
}

Mclust.machine <- function(Y.train, Y.test, mod=NULL, param = NULL, ncls, nboot = 100, alpha = 0.05, method='norm'){
  Y.train <- Y.train 
  Y.test <- Y.test  
  P <- dim(Y.train)[2]
  if (is.null(dim(Y.test)) == TRUE){
    test <- matrix(Y.test, 1, P, byrow = TRUE)
    n <- 1
  } else {
    n <- dim(Y.test)[1] 
  }
  n.tr <- dim(Y.train)[1]
  nboot <- nboot
  alpha <- alpha
  ncls <- ncls
  method <- method
  
  if (is.null(param) == TRUE){
    # Train the model #
    mod <- mod # mod must be the object of Mclust(data, ncls)
    if (is.null(mod) == TRUE){
      mod <- Mclust(data = Y.train, modelName = "VVV", G = ncls)
    } else {
      mod <- mod
    }
    train.center <- mod$parameters$mean
    train.cov <- mod$parameters$variance$sigma
  } else {
    train.center <- param$mu
    train.cov <- param$cov
    mod <- NULL
  }
  
  # Calculate the stat #
  dist <- array(Y.test, c(n, P, ncls))
  tmp <- array(rep(as.vector(train.center),each = n), c(n,P,ncls)) 
  tmp <- dist - tmp
  tmp2 <- matrix(0, n, ncls) 
  for (cls in 1:ncls){
    tmp2[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
  } 
  if (method == 'norm'){
    cutoff <- matrix(qchisq(1-alpha,P), n, ncls)
    test.ind <- tmp2 >= cutoff  
    test.ind <- apply(test.ind, 1, prod)
  } else if (method == "sum"){
    test.stat <- apply(tmp2, 1, sum) # Take summation across the distance to the centers
  } else if (method == "min"){
    test.stat <- apply(tmp2, 1, min) # Take min across the distance to the centers
  } else if (method == "boot"){
    test.stat <- tmp2
  } 
  
  if (method %in% c("sum", "min")){
    # Generate the null distribution #
    dist <- array(Y.train, c(n.tr, P, ncls))
    tmp <- array(rep(as.vector(train.center),each = n.tr), c(n.tr,P,ncls)) 
    tmp <- dist - tmp
    tmp2 <- matrix(0, n.tr, ncls) 
    for (cls in 1:ncls){
      tmp2[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
    } 
    if (method == "sum"){
      dist <- apply(tmp2, 1, sum) # Take summation across the distance to the centers
    } else if (method == "min"){
      dist <- apply(tmp2, 1, min) # Take min across the distance to the centers
    }  
    tmp <- sample(dist, n.tr*nboot, replace = TRUE)
    boot.dist <- matrix(tmp, nboot, n.tr,  byrow = TRUE)
    boot.dist <- array(boot.dist, c(nboot, n.tr, n))
    tmp <- array(rep(test.stat, each=nboot*n.tr), c(nboot, n.tr, n))
    tmp <- boot.dist > tmp # individuals with "Large" distances will be rejected
    tmp <- apply(tmp, c(1,3), sum) / n.tr
    pvalue <- apply(tmp, 2, mean) ; rm(tmp)
    test.ind <- pvalue <= alpha # If pvalue <= alpha then reject  
  }
  
  # Make prediction #
  mclst.cls <- rep(99,n)
  if (is.null(mod) ==  FALSE){
    id <- test.ind == 0 # Individuals that were not rejected will be assigned conventionally
    if (sum(id) > 0){
      Ytmp <- matrix(Y.test[id,], sum(id), P) 
      mod1.pred <- predict(mod, Ytmp)
      mclst.cls[id] <- mod1.pred$classification
    } else {
      cat('None of observations are applicable','\n')
    } 
  } else {
    id <- test.ind == 0 # Individuals that were not rejected will be assigned conventionally
    if (sum(id) > 0){
      Ytmp <- matrix(Y.test[id,], sum(id), P)
      tmp <- matrix(0, sum(id), ncls) 
      for (cls in 1:ncls){
        tmp[,cls] <- norm.density(Y.test[id,], param$mu[,cls], param$cov[,,cls])    
      }    
      tmp <- tmp / apply(tmp, 1, sum) 
      mclst.cls[id] <- apply(tmp, 1, which.max)
    } else {
      cat('None of observations are applicable','\n')
    }
  }    
  
  output <- list()
  output$result <- test.ind # 1 if rejected
  output$mod <- mod # training model
  output$pred <- mclst.cls
  return(output)
}

df.extract <- function(object){
  a <- strsplit(object,'')[[1]][-1]
  a <- as.numeric(a)
  p <- length(a)
  tmp <- 0 
  for (k in 1:p){
    tmp <- tmp + a[k]*10^(p - k)  
  }
  return(tmp)
}
# df.extract('t100')


Mclust.machine.type1error <- function(N, param, nboot = 100, niter, alpha = 0.05, method='norm', data.dist = 'norm'){
  N <- N
  param <- param
  nboot <- nboot
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Type.1.error <- rep(0, niter)
  method <- method
  data.dist <- data.dist
  
  ## Start iteration ##
  while (iter != niter + 1){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N, N/4, replace=FALSE)
    Y.train <- Y[-test.id,]
    Y.test <- Y[test.id,]
    ntest <- dim(Y.test)[1]
    true.test <- true.cls[test.id]
    
    # Train the model # 
    if (data.dist =='norm'){
      mod <- Mclust(Y.train, modelName = "VVV", ncls) # Drop the last class as sampling bias
    } else {
      mod <- NULL
    } 
    
    # Run classification #
    if (1 == 1){ ## Use mclust regardless of data distribution 
      # Since the dataset is from MN, we fit a Normal Mixture model and use the estimated parameters. 
      out <- Mclust.machine(Y.train, Y.test, mod=mod, param = NULL, ncls, nboot = nboot, alpha = alpha, method)
    } else {
      # Since the dataset does not follow MN, we use the pre-determined mean and covariance
      out <- Mclust.machine(Y.train, Y.test, mod=NULL, param = param, ncls, nboot = nboot, alpha = alpha, method)
    }
    
    # Calculate the individuals who were rejected #
    Type.1.error[iter] <- sum(out$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
    iter <- iter + 1
  }
  return(Type.1.error)
}

Mclust.Power <- function(N, param, nboot = 100, niter, alpha = 0.05, method ='norm', data.dist = 'norm', option = 'sum'){
  N <- N
  param <- param
  nboot <- nboot
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  method <- method
  Power <- rep(NA, niter)
  data.dist <- data.dist
  
  for (iter in 1:niter){ ## HERE ##
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    ## select testing data from Class 4 ##
    test.id <- which(true.cls == ncls)
    tr.id <- which(true.cls != ncls)
    Y.train <- Y[tr.id,] # Data consist of 3-classes only
    Y.test <- Y[test.id,] # Data from 4th class
    ntest <- length(test.id)
    
    ## Now suppose that we are fitting 3-class model ##
    tmp1 <- Mclust.machine(Y.train, Y.test, mod=NULL, param = NULL, ncls-1, nboot, alpha, method)
    Power[iter] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  return(Power)
}

Mclust.machine.pred <- function(N, param, nboot = 100, niter, alpha = 0.05, method ='norm', data.dist = 'norm'){
  N <- N
  param <- param
  nboot <- nboot
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  method <- method 
  train.rate.1 <- train.rate.2 <- test.rate.1  <- pred.proposed <- overall <- pvector <- rep(0, niter)
  
  ## Start iteration ##
  while (iter != niter + 1){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N,N/4) 
    con1 <- which(true.cls != 4) # exclude cls4 from the training data
    con2 <- intersect(con1,(1:N)[-test.id]) 
    nn <- N - length(test.id) - length(con2)
    nn2 <- sample(con1, nn, replace = FALSE)
    con2 <- c(con2, nn2)
    Y.train <- Y[con2,]
    Y.test <- Y[test.id,]
    ntest <- length(test.id) 
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(Y.train, modelName = "VVV", ncls-1) # Drop the last class as sampling bias
    
    # Run classification #
    out <- Mclust.machine(Y.train, Y.test, mod=mod, param = NULL, ncls-1, nboot = nboot, alpha = 0.05, method)
    pred1 <- out$pred # proposed method
    pred2 <- predict(mod, Y.test) 
    pred2 <- pred2$classification # conventional method
    
    # Seperate true classes #
    id <- true.test == ncls # individuals in the testing classes
    p <- 1 - sum(id)/ntest # Proportion of individuals in the training classes
    true.test1 <- true.test[!id] # True class memberships in the training classes
    true.test2 <- true.test[id] # True class memberships in the testing classes
    
    # Calculate the prediction rate #
    train.rate.1[iter] <- Rand.index(pred1[!id], true.test1)
    train.rate.2[iter] <- Rand.index(pred2[!id], true.test1) # Overall prediction rate of conventional method
    test.rate.1[iter] <- sum(pred1[id] == 99) / sum(id) # Proportion of correct decision in the testing class
    pred.proposed[iter] <- train.rate.1[iter]*p + test.rate.1[iter]*(1-p) # Overall prediction rate
    overall[iter] <- pred.proposed[iter] - train.rate.2[iter]*p 
    pvector[iter] <- p # Proportion of individuals in the training classes
    cat('...',iter,'th iteration is done','\n')
    iter <- iter + 1
  }
  pred.rate <- cbind(train.rate.1, train.rate.2, test.rate.1, pred.proposed, overall, pvector)
  colnames(pred.rate) <- c("Proposed.tr","Conventional.tr","Proposed.tst", "Proposed.overall", "overall", "Pvector") 
  return(pred.rate)
}
## Power comparison between others ##

knn.dist <- function(test.obs, dat, K = 10){
  N <- dim(dat)[1]
  P <- dim(dat)[2]
  test.obs <- test.obs
  test.mat <- matrix(test.obs, N, P, byrow = TRUE)
  knn <- diag((dat - test.mat) %*% t(dat - test.mat)) ## Euclidian distance
  cutoff <- max(knn[rank(knn) <= K])
  return(cutoff)
}

ODD.machine <- function(Y.train, Y.test, mod = NULL, param = NULL, ncls, alpha = 0.05, option){
  Y.train <- Y.train 
  Y.test <- Y.test  
  P <- dim(Y.train)[2]
  if (is.null(dim(Y.test)) == TRUE){
    test <- matrix(Y.test, 1, P, byrow = TRUE)
    n <- 1
  } else {
    n <- dim(Y.test)[1] 
  }
  n.tr <- dim(Y.train)[1]
  alpha <- alpha
  ncls <- ncls
  option <- option
  mod <- mod ## NULL or mclust object
  
  ## Identify the classifier ##
  if (is.null(mod) == TRUE){
    train.center <- param$mu
    train.cov <- param$cov
    mod <- NULL
  } else {
    train.center <- mod$parameters$mean
    train.cov <- mod$parameters$variance$sigma
  }
  
  ## Calculate the stat ##
  dist <- array(Y.test, c(n, P, ncls))
  tmp <- array(rep(as.vector(train.center),each = n), c(n,P,ncls)) 
  tmp <- dist - tmp
  tmp2 <- matrix(0, n, ncls) 
  
  if (option =='mahalan'){
    for (cls in 1:ncls){
      tmp2[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
    }
    dist.tr <- array(Y.train, c(n.tr, P, ncls))
    tmp <- array(rep(as.vector(train.center), each = n.tr), c(n.tr, P, ncls)) 
    tmp <- dist.tr - tmp
    tmp.tr <- matrix(0, n.tr, ncls) 
    for (cls in 1:ncls){
      tmp.tr[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
    }
    cutoff <- matrix(unname(quantile(apply(tmp.tr, 1, min), 1-alpha)), n, ncls)
  } else if (option =='euclidean'){
    for (cls in 1:ncls){
      tmp2[,cls] <- diag(tmp[,,cls] %*% t(tmp[,,cls])) # Euclidean distance
    }
    dist.tr <- array(Y.train, c(n.tr, P, ncls))
    tmp <- array(rep(as.vector(train.center), each = n.tr), c(n.tr, P, ncls)) 
    tmp <- dist.tr - tmp
    tmp.tr <- matrix(0, n.tr, ncls) 
    for (cls in 1:ncls){
      tmp.tr[,cls] <- diag(tmp[,,cls] %*% t(tmp[,,cls])) # Euclidean distance
    }
    cutoff <- matrix(unname(quantile(apply(tmp.tr, 1, min), 1-alpha)), n, ncls)
  } else if (option =='manhattan'){
    for (cls in 1:ncls){
      tmp2[,cls] <- apply(abs(tmp[,,cls]), 1, sum) # Manhattan distance  
    }
    dist.tr <- array(Y.train, c(n.tr, P, ncls))
    tmp <- array(rep(as.vector(train.center), each = n.tr), c(n.tr, P, ncls)) 
    tmp <- dist.tr - tmp
    tmp.tr <- matrix(0, n.tr, ncls) 
    for (cls in 1:ncls){
      tmp.tr[,cls] <- apply(abs(tmp[,,cls]), 1, sum) # Manhattan distance
    }
    cutoff <- matrix(unname(quantile(apply(tmp.tr, 1, min), 1-alpha)), n, ncls)
  } else if (option == 'norm'){
    for (cls in 1:ncls){
      tmp2[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
    }
    cutoff <- matrix(qchisq(1-alpha,P), n, ncls) # Chisq quantile
  } else if (option == 'weighted'){
    dist.tr <- array(Y.train, c(n.tr, P, ncls))
    tmp <- array(rep(as.vector(train.center), each = n.tr), c(n.tr, P, ncls)) 
    tmp <- dist.tr - tmp
    tmp.tr <- matrix(0, n.tr, ncls) 
    for (cls in 1:ncls){
      tmp.tr[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
    }
    cls.membership <- apply(tmp.tr, 1, which.min)
    cls.size <- rep(NA, ncls)
    for (cls in 1:ncls){
      cls.size[cls] <- sum(cls.membership == cls)
    } 
    tmp.tr <- tmp.tr * matrix(cls.size, n.tr, ncls, byrow = TRUE) # Weighted distance
    cutoff <- matrix(unname(quantile(apply(tmp.tr, 1, min), 1-alpha)), n, ncls)
    dist <- array(Y.test, c(n, P, ncls))
    tmp <- array(rep(as.vector(train.center),each = n), c(n,P,ncls)) 
    tmp <- dist - tmp
    tmp2 <- matrix(0, n, ncls) 
    for (cls in 1:ncls){
      tmp2[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
    }
    tmp2 <- tmp2 * matrix(cls.size, n, ncls, byrow = TRUE)
  } else if (option =='knn'){
    distances <- apply(Y.train, 1, knn.dist, dat = Y.train, K = 10)
    cutoff <- unname(quantile(distances, 1-alpha))
  } 
  if (option =='knn'){
    Y.tests.cut <- apply(Y.test, 1, knn.dist, dat = Y.train, K = 10) # Find the cutoff value based on Y.train
    test.ind <- Y.tests.cut >= cutoff 
  } else {
    test.ind <- tmp2 >= cutoff 
    test.ind <- apply(test.ind, 1, prod)
  }
  
  ## Make prediction ##
  mclst.cls <- rep(99,n)
  id <- test.ind == 0 # Individuals that were rejected will be assigned conventionally
  if (sum(id) > 0){
    nn <- sum(id)
    Ytmp <- matrix(Y.test[id,], nn, P) 
    if (is.null(mod) == TRUE){
      ## Calculate distances ##
      dist <- array(Ytmp, c(sum(id), P, ncls))
      tmp <- array(rep(as.vector(train.center), each = nn), c(nn,P,ncls)) 
      tmp <- dist - tmp
      tmp2 <- matrix(0, nn, ncls) 
      for (cls in 1:ncls){
        tmp2[,cls] <- diag(tmp[,,cls] %*% solve(train.cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
      } 
      mclst.cls[id] <- apply(tmp2, 1, which.min)
    } else {
      mod.pred <- predict(mod, Ytmp)
      mclst.cls[id] <- mod.pred$classification
    }
  } else {
    cat('None of observations are applicable','\n')
  }   
  
  output <- list()
  output$result <- test.ind # 1 if rejected
  output$mod <- mod # training model
  output$pred <- mclst.cls
  return(output)
}

ODD.TypeI.compare <- function(N, param, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Power <- matrix(NA, niter, 7)
  
  for (iter in 1:niter){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N, N/4, replace=FALSE)
    Y.train <- Y[-test.id,]
    Y.test <- Y[test.id,]
    ntest <- dim(Y.test)[1]
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(data = Y.train, modelName = "VVV", G = ncls)
    
    ## Now suppose that we are fitting 4-class model ##
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, alpha = alpha, option = 'norm')
    Power[iter,1] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, alpha = alpha, option = 'mahalan')
    Power[iter,2] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, alpha = alpha, option = 'manhattan')
    Power[iter,3] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, alpha = alpha, option = 'euclidean')
    Power[iter,4] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, alpha = alpha, option = 'weighted')
    Power[iter,5] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, alpha = alpha, option = 'knn')
    Power[iter,6] <- sum(tmp1$result == 1) / ntest
    tmp1 <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls, nboot = 100, alpha = alpha, method='min')
    Power[iter,7] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  colnames(Power) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap')
  return(Power)
}



ODD.TypeI.compare.unspecified <- function(N, param, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Power <- matrix(NA, niter, 7)
  ncls.vector <- rep(0, niter)
  
  for (iter in 1:niter){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N, N/4, replace=FALSE)
    Y.train <- Y[-test.id,]
    Y.test <- Y[test.id,]
    ntest <- dim(Y.test)[1]
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(data = Y.train, modelName = "VVV")
    ncls.specified <- mod$G 
    ncls.vector[iter] <- ncls.specified
    
    ## Now suppose that we are fitting 4-class model ##
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, alpha = alpha, option = 'norm')
    Power[iter,1] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, alpha = alpha, option = 'mahalan')
    Power[iter,2] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, alpha = alpha, option = 'manhattan')
    Power[iter,3] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, alpha = alpha, option = 'euclidean')
    Power[iter,4] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, alpha = alpha, option = 'weighted')
    Power[iter,5] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, alpha = alpha, option = 'knn')
    Power[iter,6] <- sum(tmp1$result == 1) / ntest
    tmp1 <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.specified, nboot = 100, alpha = alpha, method='min')
    Power[iter,7] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  colnames(Power) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap')
  output <- list()
  output$Power <- Power
  output$ncls.vector <- ncls.vector
  return(output)
}



ODD.TypeI.compare.misspecified <- function(N, param, ncls.speficied, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Power <- matrix(NA, niter, 7)
  ncls.speficied <- ncls.speficied
  
  for (iter in 1:niter){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N, N/4, replace=FALSE)
    Y.train <- Y[-test.id,]
    Y.test <- Y[test.id,]
    ntest <- dim(Y.test)[1]
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(data = Y.train, modelName = "VVV", G = ncls.speficied)
    
    ## Now suppose that we are fitting 4-class model ##
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, alpha = alpha, option = 'norm')
    Power[iter,1] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, alpha = alpha, option = 'mahalan')
    Power[iter,2] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, alpha = alpha, option = 'manhattan')
    Power[iter,3] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, alpha = alpha, option = 'euclidean')
    Power[iter,4] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, alpha = alpha, option = 'weighted')
    Power[iter,5] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, alpha = alpha, option = 'knn')
    Power[iter,6] <- sum(tmp1$result == 1) / ntest
    tmp1 <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls = ncls.speficied, nboot = 100, alpha = alpha, method='min')
    Power[iter,7] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  colnames(Power) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap')
  return(Power)
}


ODD.Power.compare <- function(N, param, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Power <- matrix(NA, niter, 7)
  
  for (iter in 1:niter){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    ## Select testing data from Class 4 ##
    test.id <- which(true.cls == ncls)
    tr.id <- which(true.cls != ncls)
    Y.train <- Y[tr.id,] # Data consist of 3-classes only
    Y.test <- Y[test.id,] # Data from 4th class
    ntest <- length(test.id)
    
    ## Train the model ## 
    mod <- Mclust(data = Y.train, modelName = "VVV", G = ncls - 1)
    
    ## Now suppose that we are fitting 3-class model ##
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'norm')
    Power[iter,1] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'mahalan')
    Power[iter,2] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'manhattan')
    Power[iter,3] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'euclidean')
    Power[iter,4] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'weighted')
    Power[iter,5] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'knn')
    Power[iter,6] <- sum(tmp1$result == 1) / ntest
    tmp1 <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, nboot = 100, alpha = alpha, method='min')
    Power[iter,7] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  colnames(Power) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap')
  return(Power)
}

ODD.Power.compare.unspecified <- function(N, param, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Power <- matrix(NA, niter, 7)
  ncls.vector <- rep(0, niter)
  
  for (iter in 1:niter){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    ## Select testing data from Class 4 ##
    test.id <- which(true.cls == ncls)
    tr.id <- which(true.cls != ncls)
    Y.train <- Y[tr.id,] # Data consist of 3-classes only
    Y.test <- Y[test.id,] # Data from 4th class
    ntest <- length(test.id)
    
    ## Train the model ## 
    mod <- Mclust(data = Y.train, modelName = "VVV")
    ncls.specified <- mod$G 
    ncls.vector[iter] <- ncls.specified
    
    ## Now suppose that we are fitting ncls.speficied-class model ##
    ## In practice, we should fit ncls = 3 but we misspecify thus ncls.specified##     
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'norm')
    Power[iter,1] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'mahalan')
    Power[iter,2] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'manhattan')
    Power[iter,3] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'euclidean')
    Power[iter,4] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'weighted')
    Power[iter,5] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'knn')
    Power[iter,6] <- sum(tmp1$result == 1) / ntest
    tmp1 <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, nboot = 100, alpha = alpha, method='min')
    Power[iter,7] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  colnames(Power) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap')
  output <- list()
  output$Power <- Power
  output$ncls.vector <- ncls.vector
  return(output)
} 


ODD.Power.compare.misspecified <- function(N, param, ncls.speficied, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  Power <- matrix(NA, niter, 7)
  ncls.speficied <- ncls.speficied
  
  for (iter in 1:niter){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    ## Select testing data from Class 4 ##
    test.id <- which(true.cls == ncls)
    tr.id <- which(true.cls != ncls)
    Y.train <- Y[tr.id,] # Data consist of 3-classes only
    Y.test <- Y[test.id,] # Data from 4th class
    ntest <- length(test.id)
    
    ## Train the model ## 
    mod <- Mclust(data = Y.train, modelName = "VVV", G = ncls.speficied)
    
    ## Now suppose that we are fitting ncls.speficied-class model ##
    ## In practice, we should fit ncls = 3 but we misspecify thus ncls.speficied ##     
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, alpha = alpha, option = 'norm')
    Power[iter,1] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, alpha = alpha, option = 'mahalan')
    Power[iter,2] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, alpha = alpha, option = 'manhattan')
    Power[iter,3] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, alpha = alpha, option = 'euclidean')
    Power[iter,4] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, alpha = alpha, option = 'weighted')
    Power[iter,5] <- sum(tmp1$result == 1) / ntest
    tmp1 <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, alpha = alpha, option = 'knn')
    Power[iter,6] <- sum(tmp1$result == 1) / ntest
    tmp1 <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.speficied, nboot = 100, alpha = alpha, method='min')
    Power[iter,7] <- sum(tmp1$result == 1) / ntest
    cat('...',iter,'th iteration is done','\n')
  }
  colnames(Power) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap')
  return(Power)
} 

ODD.pred.compare <- function(N, param, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  rate.covered <- rate.uncovered <- pred.mat <- matrix(NA, niter, 8)
  
  ## Start iteration ##
  while (iter != niter + 1){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N,N/4) 
    con1 <- which(true.cls != 4) # exclude cls4 from the training data
    con2 <- intersect(con1,(1:N)[-test.id]) 
    nn <- N - length(test.id) - length(con2)
    nn2 <- sample(con1, nn, replace = FALSE)
    con2 <- c(con2, nn2)
    Y.train <- Y[con2,]
    Y.test <- Y[test.id,]
    ntest <- length(test.id) 
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(Y.train, modelName = "VVV", ncls-1) # Drop the last class as sampling bias
    
    # Run classification #
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'norm')
    pred.1 <- out$pred # 'norm'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'mahalan')
    pred.2 <- out$pred # 'mahalan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'manhattan')
    pred.3 <- out$pred # 'manhattan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'euclidean')
    pred.4 <- out$pred # 'euclidean'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'weighted')
    pred.5 <- out$pred # 'weighted'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, alpha = alpha, option = 'knn')
    pred.6 <- out$pred # 'knn'
    out <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls - 1, nboot = 100, alpha = alpha, method='min')
    pred.7 <- out$pred # 'bootstrap'
    out <- predict(mod, Y.test) 
    pred.8 <- out$classification # conventional method
    
    # Seperate true classes #
    id <- true.test == ncls # individuals in the testing classes
    p <- 1 - sum(id)/ntest # Proportion of individuals in the training classes
    true.covered <- true.test[!id] # True class memberships in the training classes
    true.uncovered <- true.test[id] # True class memberships in the testing classes
    
    # Calculate the prediction rate #
    rate.covered[iter,1] <- Rand.index(pred.1[!id], true.covered)
    rate.uncovered[iter,1] <- sum(pred.1[id] == 99) / sum(id) # Proportion of correct decision in the testing class
    rate.covered[iter,2] <- Rand.index(pred.2[!id], true.covered)
    rate.uncovered[iter,2] <- sum(pred.2[id] == 99) / sum(id) 
    rate.covered[iter,3] <- Rand.index(pred.3[!id], true.covered)
    rate.uncovered[iter,3] <- sum(pred.3[id] == 99) / sum(id) 
    rate.covered[iter,4] <- Rand.index(pred.4[!id], true.covered)
    rate.uncovered[iter,4] <- sum(pred.4[id] == 99) / sum(id) 
    rate.covered[iter,5] <- Rand.index(pred.5[!id], true.covered)
    rate.uncovered[iter,5] <- sum(pred.5[id] == 99) / sum(id) 
    rate.covered[iter,6] <- Rand.index(pred.6[!id], true.covered)
    rate.uncovered[iter,6] <- sum(pred.6[id] == 99) / sum(id)
    rate.covered[iter,7] <- Rand.index(pred.7[!id], true.covered)
    rate.uncovered[iter,7] <- sum(pred.7[id] == 99) / sum(id)
    rate.covered[iter,8] <- Rand.index(pred.8[!id], true.covered)
    rate.uncovered[iter,8] <- 0 # The conventional method cannot digonoise
    pred.mat[iter,] <- rate.covered[iter,]*p + rate.uncovered[iter,]*(1 - p) # Overall prediction rate
    cat('...',iter,'th iteration is done','\n')
    iter <- iter + 1
  }
  output <- list()
  colnames(rate.uncovered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(rate.covered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(pred.mat) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  output$pred.mat <- pred.mat
  output$rate.covered <- rate.covered
  output$rate.uncovered <- rate.uncovered
  return(output)
}

ODD.pred.compare2 <- function(N, param, alpha = 0.05, niter = 100, classifier = 'norm', data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  classifier <- classifier # 'norm' or 'none'
  rate.covered <- rate.uncovered <- pred.mat <- matrix(NA, niter, 8)
  
  ## Start iteration ##
  while (iter != niter + 1){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N,N/4) 
    con1 <- which(true.cls != 4) # exclude cls4 from the training data
    con2 <- intersect(con1,(1:N)[-test.id]) 
    nn <- N - length(test.id) - length(con2)
    nn2 <- sample(con1, nn, replace = FALSE)
    con2 <- c(con2, nn2)
    Y.train <- Y[con2,]
    Y.test <- Y[test.id,]
    ntest <- length(test.id) 
    true.test <- true.cls[test.id]
    true.train <- true.cls[con2]
    
    # Train the model # 
    if (classifier == 'norm'){
      mod <- Mclust(Y.train, modelName = "VVV", ncls-1) # Drop the last class as sampling bias
      ini.val <- NULL 
    } else {
      mod <- NULL
      ini.val <- list()
      ini.val$mu <- matrix(0,P,ncls - 1) 
      ini.val$cov <- array(0,c(P,P,ncls - 1)) 
      for (cls in 1:(ncls - 1)){
        id <- which(true.train == cls)   
        ini.val$mu[,cls] <- apply(Y.train[id,], 2, mean, na.rm = TRUE)
        ini.val$cov[,,cls] <- cov(Y.train[id,])
      }
    }
    
    # Run classification #
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, alpha = alpha, option = 'norm')
    pred.1 <- out$pred # 'norm'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, alpha = alpha, option = 'mahalan')
    pred.2 <- out$pred # 'mahalan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, alpha = alpha, option = 'manhattan')
    pred.3 <- out$pred # 'manhattan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, alpha = alpha, option = 'euclidean')
    pred.4 <- out$pred # 'euclidean'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, alpha = alpha, option = 'weighted')
    pred.5 <- out$pred # 'weighted'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, alpha = alpha, option = 'knn')
    pred.6 <- out$pred # 'knn'
    out <- Mclust.machine(Y.train, Y.test, mod = mod, param = ini.val, ncls - 1, nboot = 100, alpha = alpha, method='min')
    pred.7 <- out$pred # 'bootstrap'
    if (classifier == 'norm'){
      out <- predict(mod, Y.test) 
      pred.8 <- out$classification # conventional method
    } else {
      nn <- dim(Y.test)[1] 
      dist <- array(Y.test, c(nn, P, ncls - 1))
      tmp <- array(rep(as.vector(ini.val$mu), each = nn), c(nn,P,ncls - 1)) 
      tmp <- dist - tmp
      tmp2 <- matrix(0, nn, ncls - 1) 
      for (cls in 1:(ncls - 1)){
        tmp2[,cls] <- diag(tmp[,,cls] %*% solve(ini.val$cov[,,cls]) %*% t(tmp[,,cls])) # Mahalanobis distance
      } 
      pred.8 <- apply(tmp2, 1, which.min)
    } 
    
    # Seperate true classes #
    id <- true.test == ncls # individuals in the testing classes
    p <- 1 - sum(id)/ntest # Proportion of individuals in the training classes
    true.covered <- true.test[!id] # True class memberships in the training classes
    true.uncovered <- true.test[id] # True class memberships in the testing classes
    
    # Calculate the prediction rate #
    rate.covered[iter,1] <- Rand.index(pred.1[!id], true.covered)
    rate.uncovered[iter,1] <- sum(pred.1[id] == 99) / sum(id) # Proportion of correct decision in the testing class
    rate.covered[iter,2] <- Rand.index(pred.2[!id], true.covered)
    rate.uncovered[iter,2] <- sum(pred.2[id] == 99) / sum(id) 
    rate.covered[iter,3] <- Rand.index(pred.3[!id], true.covered)
    rate.uncovered[iter,3] <- sum(pred.3[id] == 99) / sum(id) 
    rate.covered[iter,4] <- Rand.index(pred.4[!id], true.covered)
    rate.uncovered[iter,4] <- sum(pred.4[id] == 99) / sum(id) 
    rate.covered[iter,5] <- Rand.index(pred.5[!id], true.covered)
    rate.uncovered[iter,5] <- sum(pred.5[id] == 99) / sum(id) 
    rate.covered[iter,6] <- Rand.index(pred.6[!id], true.covered)
    rate.uncovered[iter,6] <- sum(pred.6[id] == 99) / sum(id)
    rate.covered[iter,7] <- Rand.index(pred.7[!id], true.covered)
    rate.uncovered[iter,7] <- sum(pred.7[id] == 99) / sum(id)
    rate.covered[iter,8] <- Rand.index(pred.8[!id], true.covered)
    rate.uncovered[iter,8] <- 0 # The conventional method cannot digonoise
    pred.mat[iter,] <- rate.covered[iter,]*p + rate.uncovered[iter,]*(1 - p) # Overall prediction rate
    cat('...',iter,'th iteration is done','\n')
    iter <- iter + 1
  }
  output <- list()
  colnames(rate.uncovered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(rate.covered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(pred.mat) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  output$pred.mat <- pred.mat
  output$rate.covered <- rate.covered
  output$rate.uncovered <- rate.uncovered
  return(output)
}


ODD.pred.compare.misspecified <- function(N, param, ncls.specified, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  ncls.specified <- ncls.specified
  rate.covered <- rate.uncovered <- pred.mat <- matrix(NA, niter, 8)
  
  ## Start iteration ##
  while (iter != niter + 1){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N,N/4) 
    con1 <- which(true.cls != 4) # exclude cls4 from the training data
    con2 <- intersect(con1,(1:N)[-test.id]) 
    nn <- N - length(test.id) - length(con2)
    nn2 <- sample(con1, nn, replace = FALSE)
    con2 <- c(con2, nn2)
    Y.train <- Y[con2,]
    Y.test <- Y[test.id,]
    ntest <- length(test.id) 
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(Y.train, modelName = "VVV", ncls.specified) # possibly misspecified number of cluster
    
    # Run classification #
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'norm')
    pred.1 <- out$pred # 'norm'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'mahalan')
    pred.2 <- out$pred # 'mahalan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'manhattan')
    pred.3 <- out$pred # 'manhattan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'euclidean')
    pred.4 <- out$pred # 'euclidean'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'weighted')
    pred.5 <- out$pred # 'weighted'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'knn')
    pred.6 <- out$pred # 'knn'
    out <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, nboot = 100, alpha = alpha, method='min')
    pred.7 <- out$pred # 'bootstrap'
    out <- predict(mod, Y.test) 
    pred.8 <- out$classification # conventional method
    
    # Seperate true classes #
    id <- true.test == ncls # individuals in the testing classes
    p <- 1 - sum(id)/ntest # Proportion of individuals in the training classes
    true.covered <- true.test[!id] # True class memberships in the training classes
    true.uncovered <- true.test[id] # True class memberships in the testing classes
    
    # Calculate the prediction rate #
    rate.covered[iter,1] <- Rand.index(pred.1[!id], true.covered)
    rate.uncovered[iter,1] <- sum(pred.1[id] == 99) / sum(id) # Proportion of correct decision in the testing class
    rate.covered[iter,2] <- Rand.index(pred.2[!id], true.covered)
    rate.uncovered[iter,2] <- sum(pred.2[id] == 99) / sum(id) 
    rate.covered[iter,3] <- Rand.index(pred.3[!id], true.covered)
    rate.uncovered[iter,3] <- sum(pred.3[id] == 99) / sum(id) 
    rate.covered[iter,4] <- Rand.index(pred.4[!id], true.covered)
    rate.uncovered[iter,4] <- sum(pred.4[id] == 99) / sum(id) 
    rate.covered[iter,5] <- Rand.index(pred.5[!id], true.covered)
    rate.uncovered[iter,5] <- sum(pred.5[id] == 99) / sum(id) 
    rate.covered[iter,6] <- Rand.index(pred.6[!id], true.covered)
    rate.uncovered[iter,6] <- sum(pred.6[id] == 99) / sum(id)
    rate.covered[iter,7] <- Rand.index(pred.7[!id], true.covered)
    rate.uncovered[iter,7] <- sum(pred.7[id] == 99) / sum(id)
    rate.covered[iter,8] <- Rand.index(pred.8[!id], true.covered)
    rate.uncovered[iter,8] <- 0 # The conventional method cannot digonoise
    pred.mat[iter,] <- rate.covered[iter,]*p + rate.uncovered[iter,]*(1 - p) # Overall prediction rate
    cat('...',iter,'th iteration is done','\n')
    iter <- iter + 1
  }
  output <- list()
  colnames(rate.uncovered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(rate.covered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(pred.mat) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  output$pred.mat <- pred.mat
  output$rate.covered <- rate.covered
  output$rate.uncovered <- rate.uncovered
  return(output)
}

ODD.pred.compare.unspecified <- function(N, param, alpha = 0.05, niter = 100, data.dist = 'norm'){
  N <- N
  param <- param
  alpha <- alpha
  ncls <- length(param$prop)
  P <- dim(param$mu)[1]
  niter <- niter
  iter <- 1
  rate.covered <- rate.uncovered <- pred.mat <- matrix(NA, niter, 8)
  ncls.vector <- rep(0,niter)
  
  ## Start iteration ##
  while (iter != niter + 1){
    # Generate data #
    Y <- matrix(0,N,P)
    true.cls <- numeric(N)
    id <- rmultinom(N, 1, param$prop)
    for (cls in 1:ncls){
      tmp <- id[cls,] == 1
      if (sum(tmp) != 0){
        if (data.dist == 'norm'){
          # Generate Multivariate Normal data #
          tmp.dat <- rmvnorm(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls])
        } else {
          df <- df.extract(data.dist)
          tmp.dat <- t.density(n=sum(tmp), mean=param$mu[,cls], sigma=param$cov[,,cls], df = df)
        }
      }
      Y[tmp,] <- tmp.dat
      true.cls[tmp] <- cls
    }
    
    # Select testing data from whole populations #
    test.id <- sample(1:N,N/4) 
    con1 <- which(true.cls != 4) # exclude cls4 from the training data
    con2 <- intersect(con1,(1:N)[-test.id]) 
    nn <- N - length(test.id) - length(con2)
    nn2 <- sample(con1, nn, replace = FALSE)
    con2 <- c(con2, nn2)
    Y.train <- Y[con2,]
    Y.test <- Y[test.id,]
    ntest <- length(test.id) 
    true.test <- true.cls[test.id]
    
    # Train the model # 
    mod <- Mclust(Y.train, modelName = "VVV") # possibly misspecified number of cluster
    ncls.specified <- mod$G 
    ncls.vector[iter] <- ncls.specified
    
    # Run classification #
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'norm')
    pred.1 <- out$pred # 'norm'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'mahalan')
    pred.2 <- out$pred # 'mahalan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'manhattan')
    pred.3 <- out$pred # 'manhattan'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'euclidean')
    pred.4 <- out$pred # 'euclidean'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'weighted')
    pred.5 <- out$pred # 'weighted'
    out <- ODD.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, alpha = alpha, option = 'knn')
    pred.6 <- out$pred # 'knn'
    out <- Mclust.machine(Y.train, Y.test, mod = mod, param = NULL, ncls.specified, nboot = 100, alpha = alpha, method='min')
    pred.7 <- out$pred # 'bootstrap'
    out <- predict(mod, Y.test) 
    pred.8 <- out$classification # conventional method
    
    # Seperate true classes #
    id <- true.test == ncls # individuals in the testing classes
    p <- 1 - sum(id)/ntest # Proportion of individuals in the training classes
    true.covered <- true.test[!id] # True class memberships in the training classes
    true.uncovered <- true.test[id] # True class memberships in the testing classes
    
    # Calculate the prediction rate #
    rate.covered[iter,1] <- Rand.index(pred.1[!id], true.covered)
    rate.uncovered[iter,1] <- sum(pred.1[id] == 99) / sum(id) # Proportion of correct decision in the testing class
    rate.covered[iter,2] <- Rand.index(pred.2[!id], true.covered)
    rate.uncovered[iter,2] <- sum(pred.2[id] == 99) / sum(id) 
    rate.covered[iter,3] <- Rand.index(pred.3[!id], true.covered)
    rate.uncovered[iter,3] <- sum(pred.3[id] == 99) / sum(id) 
    rate.covered[iter,4] <- Rand.index(pred.4[!id], true.covered)
    rate.uncovered[iter,4] <- sum(pred.4[id] == 99) / sum(id) 
    rate.covered[iter,5] <- Rand.index(pred.5[!id], true.covered)
    rate.uncovered[iter,5] <- sum(pred.5[id] == 99) / sum(id) 
    rate.covered[iter,6] <- Rand.index(pred.6[!id], true.covered)
    rate.uncovered[iter,6] <- sum(pred.6[id] == 99) / sum(id)
    rate.covered[iter,7] <- Rand.index(pred.7[!id], true.covered)
    rate.uncovered[iter,7] <- sum(pred.7[id] == 99) / sum(id)
    rate.covered[iter,8] <- Rand.index(pred.8[!id], true.covered)
    rate.uncovered[iter,8] <- 0 # The conventional method cannot digonoise
    pred.mat[iter,] <- rate.covered[iter,]*p + rate.uncovered[iter,]*(1 - p) # Overall prediction rate
    cat('...',iter,'th iteration is done','\n')
    iter <- iter + 1
  }
  output <- list()
  colnames(rate.uncovered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(rate.covered) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  colnames(pred.mat) <- c('norm', 'mahalan', 'manhattan', 'euclidean', 'weighted', 'knn', 'Bootstrap', 'Conventional')
  output$pred.mat <- pred.mat
  output$rate.covered <- rate.covered
  output$rate.uncovered <- rate.uncovered
  output$ncls.vector <- ncls.vector
  return(output)
}

